from urwid import Button as urwid_button
from urwid import SelectableIcon as urwid_selectable
from urwid import Text
from urwid import Columns
from urwid.text_layout import calc_coords
from urwid import WidgetWrap
from urwid.widget import FLOW
from urwid.command_map import ACTIVATE
from urwid.signals import connect_signal
from urwid.split_repr import python3_repr
from urwid.util import is_mouse_press

class SelectableIconBg(Text):
    ignore_focus = False
    _selectable = True
    def __init__(self, text, cursor_position=0):
        super().__init__(text)
        self._text = text
        self._cursor_position = cursor_position

    def render(self, size, focus=False):
        if focus:
            super().set_text(("bg_inverted", self._text))
        else:
            super().set_text(self._text)
        c = super().render(size, focus)
        return c

    
    def keypress(self, size, key):
        return key

class ButtonNoArrows(WidgetWrap):
    def sizing(self):
        return frozenset([FLOW])

    button_left = Text("")
    button_right = Text("")

    signals = ["click"]

    def __init__(self, label, on_press=None, user_data=None):
        """
        :param label: markup for button label
        :param on_press: shorthand for connect_signal()
                         function call for a single callback
        :param user_data: user_data for on_press
        Signals supported: ``'click'``
        Register signal handler with::
          urwid.connect_signal(button, 'click', callback, user_data)
        where callback is callback(button [,user_data])
        Unregister signal handlers with::
          urwid.disconnect_signal(button, 'click', callback, user_data)
        >>> Button(u"Ok")
        <Button selectable flow widget 'Ok'>
        >>> b = Button("Cancel")
        >>> b.render((15,), focus=True).text # ... = b in Python 3
        [...'< Cancel      >']
        """
        self._label = SelectableIconBg("", 0)
        cols = Columns([
            ('fixed', 1, self.button_left),
            self._label,
            ('fixed', 1, self.button_right)],
            dividechars=1)
        super().__init__(cols)

        # The old way of listening for a change was to pass the callback
        # in to the constructor.  Just convert it to the new way:
        if on_press:
            connect_signal(self, 'click', on_press, user_data)

        self.set_label(label)

    def _repr_words(self):
        # include button.label in repr(button)
        return super()._repr_words() + [
            python3_repr(self.label)]

    def set_label(self, label):
        """
        Change the button label.
        label -- markup for button label
        >>> b = Button("Ok")
        >>> b.set_label(u"Yup yup")
        >>> b
        <Button selectable flow widget 'Yup yup'>
        """
        self._label.set_text(label)

    def get_label(self):
        """
        Return label text.
        >>> b = Button(u"Ok")
        >>> print(b.get_label())
        Ok
        >>> print(b.label)
        Ok
        """
        return self._label.text
    label = property(get_label)

    def keypress(self, size, key):
        if self._command_map[key] != ACTIVATE:
            return key

        self._emit('click')

    def mouse_event(self, size, event, button, x, y, focus):
        if button != 1 or not is_mouse_press(event):
            return False

        self._emit('click')
        return True


from aiohttp import ClientSession
from jellyfin_cli.jellyfin_client.data_classes.View import View
from jellyfin_cli.jellyfin_client.data_classes.Shows import Episode, Show
from jellyfin_cli.jellyfin_client.data_classes.Movies import Movie
from jellyfin_cli.jellyfin_client.data_classes.Audio import Audio, Album

class InvalidCredentialsError(Exception):
    def __init__(self):
        Exception("Invalid username, password or server URL")

class HttpError(Exception):
    def __init__(self, text):
        Exception("Something went wrong: {}".format(text))

class ServerContext:
    def __init__(self, res=None, url=None, client=None, cfg=None, username=None):
        if cfg:
            self.url = cfg["url"]
            self.user_id = cfg["user_id"]
            self.server_id = cfg["server_id"]
            self.client = ClientSession(headers={
                "x-emby-authorization": cfg["auth_header"]
            })
            self.username = cfg["username"]
        else:
            self.url = url
            self.user_id = res["User"]["Id"]
            self.server_id = res["ServerId"]
            self.username = username
            self.client = client
    
    def get_token(self):
        header = self.client._default_headers["x-emby-authorization"]
        header = [i.split("=") for i in header.split(",")]
        pairs = {k[0].strip().replace('"',""):k[1].strip().replace('"',"") for k in header}
        return pairs["Token"]

class HttpClient:
    def __init__(self, server, context=None):
        self.client = ClientSession()
        self.server = server
        self.context = context

    async def login(self, username, password):
        try:
            res = await self.client.post(self.server+'/Users/authenticatebyname',data={
                "Username": username,
                "Pw": password
            }, headers={
                "x-emby-authorization":'MediaBrowser Client="Jellyfin CLI", Device="Jellyfin-CLI", DeviceId="None", Version="10.4.3"'
            })
        except Exception:
            raise InvalidCredentialsError()
        if res.status == 200:
            res = await res.json()
            token = res["AccessToken"]
            self.client = ClientSession(headers={
                "x-emby-authorization":'MediaBrowser Client="Jellyfin CLI", Device="Jellyfin-CLI", DeviceId="None", Version="10.4.3", Token="{}"'.format(token)
            })
            self.context = ServerContext(res, self.server, self.client, username=username)
            from jellyfin_cli.utils.login_helper import store_creds
            store_creds(self.context)
            return True
        elif res.status == 401:
            raise InvalidCredentialsError()
        else:
            raise HttpError(await res.text())
    
    async def get_views(self):
        res = await self.context.client.get("{}/Users/{}/Views".format(self.context.url, self.context.user_id))
        if res.status == 200:
            res = await res.json()
            return [View(i, self.context) for i in res["Items"]]
        else:
            raise HttpError(await res.text())

    async def get_resume(self, limit=12, types="Video"):
        res = await self.context.client.get("{}/Users/{}/Items/Resume".format(self.context.url, self.context.user_id), params={
            "Limit": limit,
            "Recursive": "true",
            "Fields": "BasicSyncInfo",
            "MediaTypes": types
        })
        if res.status == 200:
            res = await res.json()
            return [Episode(r, self.context) for r in res["Items"]]
        else:
            raise HttpError(await res.text())

    async def get_nextup(self, limit=24):
        res = await self.context.client.get("{}/Shows/NextUp".format(self.context.url), params={
            "UserId": self.context.user_id,
            "Limit": limit,
            "Recursive": "true",
            "Fields": "BasicSyncInfo"
        })
        if res.status == 200:
            res = await res.json()
            return [Episode(r, self.context) for r in res["Items"]]
        else:
            raise HttpError(await res.text())

    async def search(self, query, media_type, limit=30):
        res = await self.context.client.get("{}/Users/{}/Items".format(self.context.url, self.context.user_id), params={
            "searchTerm": query,
            "IncludeItemTypes": media_type,
            "IncludeMedia": "true",
            "IncludePeople": "false",
            "IncludeGenres": "false",
            "IncludeStudios": "false",
            "IncludeArtists": "false",
            "Fields": "BasicSyncInfo",
            "Recursive": "true",
            "Limit": limit
        })
        if res.status == 200:
            res = await res.json()
            r = []
            for i in res["Items"]:
                if i["Type"] == "Movie":
                    r.append(Movie(i, self.context))
                elif i["Type"] == "Audio":
                    r.append(Audio(i, self.context))
                elif i["Type"] == "Series":
                    r.append(Show(i, self.context))
                elif i["Type"] == "Episode":
                    r.append(Episode(i, self.context))
                elif i["Type"] == "MusicAlbum":
                    r.append(Album(i, self.context))
            return r
        else:
            raise HttpError(await res.text())

    async def get_recommended(self, limit=30):
        res = await self.context.client.get("{}/Users/{}/Items".format(self.context.url, self.context.user_id), params={
            "SortBy": "IsFavoriteOrLiked,Random",
            "IncludeItemTypes": "Movie,Series",
            "Recursive": "true",
            "Limit": limit
        })
        if res.status == 200:
            res = await res.json()
            r = []
            for i in res["Items"]:
                if i["Type"] == "Movie":
                    r.append(Movie(i, self.context))
                elif i["Type"] == "Series":
                    r.append(Show(i, self.context))
            return r
        else:
            raise HttpError(await res.text())


from jellyfin_cli.jellyfin_client.JellyfinClient import HttpClient, InvalidCredentialsError, ServerContext, HttpError
from aiohttp import ClientSession
from json import load, dump
from getpass import getpass

async def login():
    try:
        return await load_creds()
    except Exception as e:
        print(str(e))
        pass
    while True:
        server = input("Enter the server URL: ")
        if server.endswith("/"):
            server = server[:-1]
        username = input("Enter the username: ")
        password = getpass("Enter the password: ")
        client = HttpClient(server)
        try:
            await client.login(username, password)
            print("Logged in...")
            return client
        except InvalidCredentialsError as e:
            print(str(e))
            continue

def store_creds(context):
    with open("creds.json", "w") as f:
        dump({
            "url": context.url,
            "auth_header": context.client._default_headers["x-emby-authorization"],
            "user_id": context.user_id,
            "server_id": context.server_id,
            "username": context.username
        },f)

async def load_creds():
    with open("creds.json", "r") as f:
        dc = load(f)
        client = HttpClient(dc["url"], ServerContext(cfg=dc))
        await client.get_views()
        return client


from jellyfin_cli.jellyfin_client.data_classes.Items import Item
from datetime import datetime

class Movie(Item):
    def __init__(self, res, context):
        super().__init__(res, context)
        self.subbed = res["HasSubtitles"] if "HasSubtitles" in res else False
        if "PremiereDate" in res:
            self.premiered = datetime.strptime(res["PremiereDate"].split("T")[0], "%Y-%m-%d")
        else:
            self.premiered = None
        self.rating = res["OfficialRating"] if "OfficialRating" in res else None
        self.community_rating = res["CommunityRating"] if "CommunityRating" in res else None





from jellyfin_cli.jellyfin_client.data_classes.Items import Item

class Audio(Item):
    def __init__(self, res, context):
        super().__init__(res, context)
        self.production_year = res["ProductionYear"] if "ProductionYear" in res else None
        self.artists = res["Artists"] if "Artists" in res else ["Unknown Artist"]
        self.album = res["Album"] if "Album" in res else "Unknown Album"

class Album(Item):
    def __init__(self ,res, context):
        super().__init__(res, context)
        self.artists = res["Artists"]
        self.album_artist = res["AlbumArtist"] if "AlbumArtist" in res and res["AlbumArtist"] else "Unknown Artist"
        self.item_count = res["ChildCount"] if "ChildCount" in res else 0

    async def get_songs(self, sort="SortName"):
        res = await self.context.client.get("{}/Users/{}/Items".format(self.context.url, self.context.user_id), params={
            "ParentId": self.id,
            "Fields": "BasicSyncInfo",
            "SortBy": sort
        })
        if res.status == 200:
            res = await res.json()
            r = []
            for i in res["Items"]:
                r.append(Audio(i, self.context))
            return r


from jellyfin_cli.jellyfin_client.data_classes.Items import Item

class Show(Item):
    def __init__(self, res, context):
        super().__init__(res, context)
    
    async def get_seasons(self):
        res = await self.context.client.get("{}/Shows/{}/Seasons".format(self.context.url, self.id),params={
            "userId": self.context.user_id
        })
        if res.status == 200:
            res = await res.json()
            return [Season(i, self) for i in res["Items"]]
        else:
            from jellyfin_cli.jellyfin_client.JellyfinClient import HttpError
            raise HttpError(await res.text())

class Season(Item):
    def __init__(self, res, show):
        super().__init__(res, show.context)
        if "IndexNumber" in res:
            self.index = res["IndexNumber"]
        else:
            self.index = 0
        self.is_folder = res["IsFolder"]
        self.show = show
    
    def __str__(self):
        return self.name

    async def get_episodes(self):
        res = await self.context.client.get("{}/Shows/{}/Episodes".format(self.context.url, self.show.id), params={
            "seasonId": self.id,
            "userId": self.context.user_id
        })
        if res.status == 200:
            res = await res.json()
            return [Episode(i, self.context) for i in res["Items"]]
        else:
            from jellyfin_cli.jellyfin_client.JellyfinClient import HttpError
            raise HttpError(await res.text())

class Episode(Item):
    def __init__(self, res, context):
        super().__init__(res, context)
        self.subbed = res["HasSubtitles"] if "HasSubtitles" in res else False

        self.context = context

    def __str__(self):
        return "{} {}".format(self.name, "[Subtitled]" if self.subbed else "")


from datetime import datetime

class Item:
    def __init__(self, res, context):
        self.id = res["Id"]
        self.name = res["Name"]
        self.is_folder = res["IsFolder"]
        self.played = res["UserData"]["Played"]
        if "RunTimeTicks" in res:
            self.ticks = res["RunTimeTicks"]
        else:
            self.ticks = 9
        self.context = context

    def __str__(self):
        return self.name

class Playlist(Item):
    def __init__(self, res, context):
        super().__init__(res, context)

    async def get_items(self):
        res = await self.context.client.get("{}/Playlists/{}/Items".format(self.context.url, self.id), params={
            "Fields": "BasicSyncInfo",
            "UserId": self.context.user_id
        })
        if res.status == 200:
            res = await res.json()
            r = []
            from jellyfin_cli.jellyfin_client.data_classes.Audio import Audio
            from jellyfin_cli.jellyfin_client.data_classes.Shows import Episode
            from jellyfin_cli.jellyfin_client.data_classes.Movies import Movie
            for i in res["Items"]:
                if i["Type"] == "Audio":
                    r.append(Audio(i, self.context))
                elif i["Type"] == "Episode":
                    r.append(Episode(i, self.context))
                elif i["Type"] == "Movie":
                    r.append(Movie(i, self.context))
            return r
        else:
            from jellyfin_cli.jellyfin_client.JellyfinClient import HttpError
            raise HttpError(await res.text())


from os.path import isfile
from os import devnull, getenv
from aio_mpv_jsonipc import MPV
from asyncio import get_event_loop, sleep
from datetime import timedelta

def ticks_to_seconds(ticks):
    return int(ticks*(1/10000000))

def seconds_to_ticks(seconds):
    return int(seconds * pow(10, 9) / 100)

class Player:
    def __init__(self, context):
        self.context = context

        self.mpv = None
        self.item = None
        self.position = 0
        self.duration = 0
        self.playing = False
        self.paused = False

        self.played = False

    async def _get_api_keys(self):
        res = await self.context.client.get("{}/Auth/Keys".format(self.context.url))
        if res.status == 200:
            res = await res.json()
            return {i["AppName"] : i["AccessToken"] for i in res["Items"]}

    async def _get_api_key(self):
        keys = await self._get_api_keys()
        if "jellyfin_cli_play" in keys:
            return keys["jellyfin_cli_play"]
        else:    
            await self.context.client.post("{}/Auth/Keys".format(self.context.url), data={
                "App": "jellyfin_cli_play"
            })
            return await self._get_api_key()

    async def _delete_api_key(self, key=None):
        if not key:
            key = await self._get_api_key()
        await self.context.client.delete("{}/Auth/Keys/{}".format(self.context.url, key))

    
    
    async def _update_time(self, time):
        try:
            self.position = int(time)
            prcnt = (self.position/self.duration)*100
            if prcnt > 70 and not self.played:
                get_event_loop().create_task(self.context.client.post(
                    "{}/Users/{}/PlayedItems/{}".format(self.context.url, self.context.user_id, self.item.id)
                ))
                self.played = True
        except:
            pass

    async def _play(self, item, block=True):
        if self.playing:
            await self.mpv.send(["quit"])
        self.item = item
        self.position = 0
        self.duration = int(ticks_to_seconds(item.ticks))
        self.playing = True
        self.played = False
        try:
            key = await self._get_api_key()
        except:
            key = self.context.get_token()
            print("Could not create API token. I will use your login token. Be careful to not leak it!")
        url = "{}/Items/{}/Download?api_key={}".format(self.context.url, item.id, key)
        self.mpv = MPV(media=url)
        await self.mpv.start()
        self.mpv.listen_for("property-change", self._update_time)
        await self.mpv.send(["observe_property", 1, "time-pos"])
        async def _():
            await self.mpv.wait_complete()
            self.playing = False
            await self._delete_api_key(key)
        if block:
            await _()
        else:
            get_event_loop().create_task(_())

    def play(self, button, item):
        get_event_loop().create_task(self._play(item))

    def pause(self):
        if self.paused:
            self.paused = False
            get_event_loop().create_task(self.mpv.send(["set_property", "pause", False]))
        else:
            self.paused = True
            get_event_loop().create_task(self.mpv.send(["set_property", "pause", True]))

    async def stop(self):
        self.playing = False
        try:
            await self.mpv.stop()
        except:
            pass

    def get_playback_string(self):
        position = timedelta(seconds=self.position)
        duration = timedelta(seconds=self.duration)
        return " {}                     {} / {}".format(self.item.name, position, duration)


from jellyfin_cli.jellyfin_client.data_classes.Shows import Show
from jellyfin_cli.jellyfin_client.data_classes.Movies import Movie
from jellyfin_cli.jellyfin_client.data_classes.Audio import Album, Audio
from jellyfin_cli.jellyfin_client.data_classes.Items import Playlist

class HttpError(Exception):
    pass

class View:
    def __init__(self, res, context):
        self.name = res["Name"]
        self.id = res["Id"]
        self.etag = res["Etag"]
        self.parent_id = res["ParentId"] if "ParentId" in res else None
        if res["CollectionType"] == "movies":
            self.sort = "DateCreated,SortName,ProductionYear"
            self.view_type = "Movie"
        elif res["CollectionType"] == "tvshows":
            self.sort = "DateCreated,SortName,ProductionYear"
            self.view_type = "Series"
        elif res["CollectionType"] == "music":
            self.sort = "DatePlayed"
            self.view_type = "Audio"
        elif res["CollectionType"] == "playlists":
            self.sort = "IsFolder,SortName"
            self.view_type = "Playlist"
        self.context = context

    def __str__(self):
        return self.name

    async def get_items(self, start=0, limit=100, sort=None):
        if not sort:
            sort = self.sort
        res = await self.context.client.get("{}/Users/{}/Items".format(self.context.url,self.context.user_id), params={
            "SortBy": self.sort,
            "SortOrder": "Descending",
            "Recursive": "true",
            "Fields": "BasicSyncInfo",
            "StartIndex": start,
            "Limit": limit,
            "ParentId": self.id,
            "IncludeItemTypes": self.view_type
        })
        if res.status == 200:
            res = await res.json()
            r = []
            for i in res["Items"]:
                if i["Type"] == "Movie":
                    r.append(Movie(i, self.context))
                elif i["Type"] == "Audio":
                    r.append(Audio(i, self.context))
                elif i["Type"] == "Series":
                    r.append(Show(i, self.context))
                elif i["Type"] == "Playlist":
                    r.append(Playlist(i, self.context))
            return r
        else:
            raise HttpError(await res.text())

    async def get_latest(self, limit=30):
        res = await self.context.client.get("{}/Users/{}/Items/Latest".format(self.context.url, self.context.user_id), params={
            "Limit": limit,
            "ParentId": self.id,
            "Fields": "BasicSyncInfo"
        })
        if res.status == 200:
            res = await res.json()
            r = []
            for i in res:
                if i["Type"] == "Movie":
                    r.append(Movie(i, self.context))
                elif i["Type"] == "Audio":
                    r.append(Audio(i, self.context))
                elif i["Type"] == "Series":
                    r.append(Show(i, self.context))
                elif i["Type"] == "MusicAlbum":
                    r.append(Album(i, self.context))
            return r
        else:
            raise HttpError(await res.text())


from jellyfin_cli.utils.login_helper import login
from jellyfin_cli.utils.play_utils import Player
from asyncio import get_event_loop, sleep
from jellyfin_cli.urwid_overrides.Button import ButtonNoArrows as Button
import urwid

def calculate_percentage(rows, prcnt):
    return int((rows/100)*prcnt)

PALLETE = [
    ('bg', 'black', 'light gray'),
    ('bg_inverted', 'white', 'light gray')
]

class App:
    def __init__(self):
        self.loop = get_event_loop()
        self.client = self.loop.run_until_complete(login())
        self.player = Player(self.client.context)
        self.widgets = [urwid.Text("Logged in as {}".format(self.client.context.username)), urwid.Divider(bottom=1)]
        self.draw = urwid.MainLoop(urwid.Filler(urwid.Pile([])), palette=PALLETE, event_loop=urwid.AsyncioEventLoop(loop=self.loop))
        self.previous_key_callback = (None, (None))
        self.draw.unhandled_input = self._process_keypress

        self._last_view = self.draw_search
        self.search_query = ""
        self.search_edit = urwid.Edit("Enter your search query...", edit_text=self.search_query)

    def _process_keypress(self, key):
        if key == "tab":
            try:
                if self.previous_key_callback[1]:
                    self.previous_key_callback[0](*self.previous_key_callback[1])
                else:
                    self.previous_key_callback[0]()
            except:
                pass
        elif key == "s":
            self.draw_search(None, "")
        elif key == "p":
            if self.player.playing:
                self.player.pause()
        elif key == "q":
            def _(x=None):
                raise urwid.ExitMainLoop()
            if self.player.playing:
                task = self.loop.create_task(self.player.stop())
                task.add_done_callback(_)
            _()

    def _draw_table(self, items, title=None, prcnt=30, callback=None):
        if not callback:
            callback = self.play
        rows = calculate_percentage(self.draw.screen.get_cols_rows()[1], prcnt)
        texts = [Button(str(i), on_press=callback(i) if not hasattr(self, callback.__name__) else callback, user_data=i) for i in items]
        widgets = [urwid.Text(("bg", title), align=urwid.CENTER)] if title else []
        widgets.append(urwid.BoxAdapter(urwid.ListBox(urwid.SimpleFocusListWalker(texts)), rows))
        return widgets

    def _draw_screen(self):
        self.draw.widget.body = urwid.Pile(self.widgets)
    
    def _empty_screen(self, text=""):
        self.draw.widget.body = urwid.Text(text)

    def _clear_widgets(self):
        self.widgets = [urwid.Text("Logged in as {}".format(self.client.context.username)), urwid.Divider(bottom=1)]

    def _clear_search(self):
        self.search_query = ""
        self.search_edit = urwid.Edit()
        
    def _add_widget(self, wd):
        if type(wd) == list:
            self.widgets.extend(wd)
        else:
            self.widgets.append(wd)
        self.widgets.append(urwid.Divider(div_char="_", top=1, bottom=1))
        self.draw.widget.body = urwid.Pile(self.widgets)

    async def _draw_home(self):
        self.widgets = [urwid.Text("Logged in as {}".format(self.client.context.username)), urwid.Divider(bottom=1)]
        views = await self.client.get_views()
        columns = urwid.Columns([Button(("bg", i.name), on_press=self.draw_view, user_data=i) for i in views])
        self._add_widget(columns)
        resume = await self.client.get_resume()
        nextup = await self.client.get_nextup()
        self._add_widget(self._draw_table(resume,"Continue Watching", prcnt=30))
        self._add_widget(self._draw_table(nextup, "Next Up", prcnt=28))
        self._add_widget(urwid.Text("Tab: Go back | s: Search | p: Play/Pause | q: Exit"))
        self.previous_key_callback = (None, None)

    def draw_home(self):
        self.loop.create_task(self._draw_home())

    async def _draw_view(self, view):
        self._clear_widgets()
        if view.view_type == "Audio":
            await self._draw_audio(view)
        elif view.view_type == "Playlist":
            self._last_view = view
            items = await view.get_items()
            self._add_widget(self._draw_table(items, view.name, prcnt=98, callback=self.draw_playlist))
        else:
            items = await view.get_items(limit=500)
            if view.view_type == "Series":
                callback = self.draw_seasons
            else:
                callback = None
            self._add_widget(self._draw_table(items, str(view), prcnt=96, callback=callback))
            self._last_view = view
        self.previous_key_callback = (self.draw_home, None)

    def draw_view(self, b, view):
        self.loop.create_task(self._draw_view(view))

    async def _draw_playlist(self, playlist):
        items = await playlist.get_items()
        callbacks = {"Movie":self.play, "Episode": self.play, "Audio": self.play_bg}
        self._clear_widgets()
        self._add_widget(self._draw_table(items, playlist.name, prcnt=98, callback=lambda i: callbacks[i.__class__.__name__]))
        self.previous_key_callback = (self.draw_view, (None, self._last_view))
    
    def draw_playlist(self, b, playlist):
        self.loop.create_task(self._draw_playlist(playlist))

    async def _draw_audio(self, view):
        latest = await view.get_latest(limit=16)
        recent = await view.get_items()
        frequent = await view.get_items(sort="PlayCount", limit=8)
        self._add_widget(self._draw_table(latest, title="Latest Music", callback=self.draw_album, prcnt=15))
        self._add_widget(self._draw_table(recent, title="Recently Played", callback=self.play_bg, prcnt=40))
        self._add_widget(self._draw_table(frequent, title="Frequently Played", callback=self.play_bg, prcnt=15))
        self._last_view = view
        self.previous_key_callback = (self.draw_home, None)
    
    async def _draw_album(self, album):
        items = await album.get_songs()
        self._clear_widgets()
        self._add_widget(self._draw_table(items, title=album.name, prcnt=97, callback=self.play_bg))
        self.previous_key_callback = (self.draw_view, (None, self._last_view)) if not callable(self._last_view) else (self._last_view, (None, None))

    def draw_album(self, b, album):
        self.loop.create_task(self._draw_album(album))

    async def _draw_seasons(self, series):
        seasons = await series.get_seasons()
        self._clear_widgets()
        self._add_widget(self._draw_table(seasons, str(series), prcnt=96, callback=self.draw_episodes))
        self.previous_key_callback = (self.draw_view, (None, self._last_view))

    def draw_seasons(self, b, series, callback=None):
        self.loop.create_task(self._draw_seasons(series))

    async def _draw_episodes(self, season):
        episodes = await season.get_episodes()
        self._clear_widgets()
        self._add_widget(self._draw_table(episodes, "{} - {}".format(season.show, season), prcnt=96))
        self.previous_key_callback = (self.draw_seasons, (None, season.show))

    def draw_episodes(self, b, season):
        self.loop.create_task(self._draw_episodes(season))

    async def _draw_search(self):
        self._clear_widgets()
        if len(self.search_query) > 0:
            movies = await self.client.search(self.search_query, "Movie")
            shows = await self.client.search(self.search_query, "Series")
            episodes = await self.client.search(self.search_query, "Episode")
            songs = await self.client.search(self.search_query, "Audio")
            albums = await self.client.search(self.search_query, "MusicAlbum")
            self._add_widget(self.search_edit)
            if movies:
                self._add_widget(self._draw_table(movies, "Movies", callback=self.play, prcnt=20))
            if shows:
                self._add_widget(self._draw_table(shows, "Shows", callback=self.draw_seasons, prcnt=20))
            if episodes:
                self._add_widget(self._draw_table(episodes, "Episodes", callback=self.play, prcnt=20))
            if songs:
                self._add_widget(self._draw_table(songs, "Songs", callback=self.play_bg, prcnt=20))
            if albums:
                self._add_widget(self._draw_table(albums, "Albums", callback=self.draw_album, prcnt=20))
        else:
            urwid.connect_signal(self.search_edit, "change", self.draw_search)
            recommended = await self.client.get_recommended()
            self._add_widget(self.search_edit)
            callbacks = {"Movie":self.play, "Show": self.draw_seasons, "Episode": self.play}
            self._add_widget(self._draw_table(recommended, "Suggestions", prcnt=96, callback=lambda i: callbacks[i.__class__.__name__]))
        self.previous_key_callback = (lambda: [self.draw_home(), urwid.disconnect_signal(self.search_edit, "change", self.draw_search), self._clear_search()], (None))

    def draw_search(self, e, text):
        self.search_query = text if text is not None else ""
        self.loop.create_task(self._draw_search())

    async def _play(self, item):
        self._empty_screen()
        await self.player._play(item)
        self._draw_screen()

    async def _bg_play(self , item):
        await self.player._play(item, block=False)
        while self.player.playing:
            play_string = self.player.get_playback_string()
            if not self.previous_key_callback[0]:
                try:
                    self.draw.widget.body.contents[-2][0].set_text(play_string)
                    self.draw.draw_screen()
                except:
                    break
            await sleep(1)
    
    def play(self, b, item, bg=False):
        if bg:
            self.loop.create_task(self._bg_play(item))
        else:
            self.loop.create_task(self._play(item))

    def play_bg(self, b, item):
        self.play(b, item, bg=True)

    def _run(self):
        self.draw_home()
        self.draw.run()

    def __call__(self):
        return self._run()

app = App()
if __name__ == "__main__":
    app()


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='Jellyfin-CLI',
    version='1.3.1',
    scripts=['jellyfin-cli'],
    author="marios8543",
    author_email="marios8543@gmail.com",
    description="A Jellyfin command line client.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/marios8543/Jellyfin-CLI",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "jellyfin-cli = jellyfin_cli.main:app"
        ]
    },
    install_requires=[
        "aiohttp",
        "urwid",
        "aio-mpv-jsonipc"
    ]
)
